////////////////////////////////////////

// Disable macros which HPCC defines but are also typedef-ed in sqltypes.h

#pragma push_macro("CHAR")
#pragma push_macro("VOID")
#pragma push_macro("DWORD")
#pragma push_macro("LPTSTR")
#pragma push_macro("LPSTR")
#pragma push_macro("HINSTANCE")
#pragma push_macro("UINT")
#undef CHAR
#undef VOID
#undef DWORD
#undef LPTSTR
#undef LPSTR
#undef HINSTANCE
#undef UINT

#include <cstdio>
#include <sql.h>
#include <sqlext.h>
#include <cstring>
#include <string>
#include <vector>
#include <iostream>

// Restore the HPCC macro definitions
#pragma pop_macro("CHAR")
#pragma pop_macro("VOID")
#pragma pop_macro("DWORD")
#pragma pop_macro("LPTSTR")
#pragma pop_macro("LPSTR")
#pragma pop_macro("HINSTANCE")
#pragma pop_macro("UINT")

using namespace std;
////////////////////////////////////////

// Diagnostic function
void extract_error(
    const string fn,
    SQLHANDLE handle,
    SQLSMALLINT type)
{
    SQLINTEGER	 i = 0;
    SQLINTEGER	 native;
    SQLCHAR	 state[ 7 ];
    SQLCHAR	 text[256];
    SQLSMALLINT	 len;
    SQLRETURN	 ret;
    cerr << endl << "The driver reported the following diagnostics whilst running " << fn << endl;
    do
    {
        ret = SQLGetDiagRec(type, handle, ++i, state, &native, text,
                            sizeof(text), &len );
        if (SQL_SUCCEEDED(ret))
            printf("%s:%ld:%ld:%s\n", state, i, native, text);
    }
    while( ret == SQL_SUCCESS );
}

// Define The ODBC_Class Class
class ODBC_Class
{
	struct driverDescription
	{
		SQLCHAR dbms_name[256];
		SQLCHAR dbms_ver[256];
		SQLUINTEGER getdata_support;
		SQLUSMALLINT max_concur_act;
		SQLSMALLINT string_len;
	} driverInfo;
// Attributes
public:
	SQLHENV envHandle;
	SQLHDBC dbcHandle;
	SQLHSTMT stmtHandle;
	SQLRETURN rc;
// Operations
public:
	ODBC_Class(const char *DSN); // Constructor
	~ODBC_Class(); // Destructor
	void getDriverInfo();
private:
};

// Define The Class Constructor
ODBC_Class::ODBC_Class(const char *DSN)
{
	// Allocate an environment handle
	rc = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &envHandle);
	if(!SQL_SUCCEEDED(rc)) {
		cerr << "Could not allocate environment handle" << endl;
		return;
	}
	// Set The ODBC Application Version To 3.x
	rc = SQLSetEnvAttr(envHandle, SQL_ATTR_ODBC_VERSION,
		(SQLPOINTER) SQL_OV_ODBC3, SQL_IS_UINTEGER);
	if(!SQL_SUCCEEDED(rc)) {
		cerr << "Could not set version attribute" << endl;
		return;
	}
	// Allocate a database connection handle
	rc = SQLAllocHandle(SQL_HANDLE_DBC, envHandle, &dbcHandle);
	if(!SQL_SUCCEEDED(rc)) {
		cerr << "Could not allocate connection handle" << endl;
		return;
	}
	// Connect to the data source	
	rc = SQLConnect(dbcHandle,(SQLCHAR *)DSN,SQL_NTS,NULL,0,NULL,0);
	if(!SQL_SUCCEEDED(rc)) {
		cerr << "Could not connect to database" << endl;
		extract_error("SQLConnect",dbcHandle,SQL_HANDLE_DBC);
		return;
	}
	// Allocate a SQL statement handle
	rc = SQLAllocHandle(SQL_HANDLE_STMT,dbcHandle,&stmtHandle);
	if(!SQL_SUCCEEDED(rc)) {
		cerr << "Could not allocate statement handle" << endl;
		return;
	}
	// Get some driver information
	SQLGetInfo(dbcHandle,SQL_DBMS_NAME,(SQLPOINTER)driverInfo.dbms_name,sizeof(driverInfo.dbms_name),NULL);
	SQLGetInfo(dbcHandle,SQL_DBMS_VER,(SQLPOINTER)driverInfo.dbms_ver,sizeof(driverInfo.dbms_ver),NULL);
	SQLGetInfo(dbcHandle,SQL_GETDATA_EXTENSIONS,(SQLPOINTER)&(driverInfo.getdata_support),0,0);
	SQLGetInfo(dbcHandle,SQL_MAX_CONCURRENT_ACTIVITIES,&(driverInfo.max_concur_act),0,0);
}

// Define The Class Destructor
ODBC_Class::~ODBC_Class()
{
	// Free the statement handle
	if (stmtHandle != NULL) {
		rc = SQLFreeHandle(SQL_HANDLE_STMT, stmtHandle); 
	    if(!SQL_SUCCEEDED(rc)) {
		    cerr << "Could not free statement handle" << endl;
            extract_error("SQLFreeHandle",stmtHandle,SQL_HANDLE_STMT);
	    }
    }
	// Free the database connection handle
	if (dbcHandle != NULL) {
	    // Disconnect from the data source
	    rc = SQLDisconnect(dbcHandle);
        if(!SQL_SUCCEEDED(rc)) {
            cerr << "Could not disconnect" << endl;
            extract_error("SQLDisconnect",dbcHandle,SQL_HANDLE_DBC);
        }
        // FIXME: Freeing connection handle when pooling is enabled sometimes (HThor but not Roxie) leads to a locked state.
		rc = SQLFreeHandle(SQL_HANDLE_DBC, dbcHandle);
	    if(!SQL_SUCCEEDED(rc)) {
		    cerr << "Could not free connection handle" << endl;
            extract_error("SQLFreeHandle",dbcHandle,SQL_HANDLE_DBC);
	    }
    }
	// Free the environment handle
	if (envHandle != NULL) {
		rc = SQLFreeHandle(SQL_HANDLE_ENV, envHandle);
	    if(!SQL_SUCCEEDED(rc)) {
		    cerr << "Could not free environment handle" << endl;
            extract_error("SQLFreeHandle",envHandle,SQL_HANDLE_ENV);
	    }
    }
}

void 
ODBC_Class::getDriverInfo()
{
	cout << "DBMS Name: " << driverInfo.dbms_name << endl;
	cout << "DBMS Version: " << driverInfo.dbms_ver << endl;
	cout << "Maximum Concurrent Threads: " << driverInfo.max_concur_act << endl;
	if (driverInfo.getdata_support & SQL_GD_ANY_ORDER)
		cout << "SQLGetData - columns can be retrieved in any order" << endl;
	else 
		cout << "SQLGetData - columns must be retrieved in order" << endl;
	if (driverInfo.getdata_support & SQL_GD_ANY_COLUMN)
		cout << "SQLGetData - can retrieve columns before last bound one" << endl;
	else 
		cout << "SQLGetData - must retrieve columns after last bound one" << endl;
}

